#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>

#include "SDL.h"
#define NO_SDL_GLEXT
#include <SDL/SDL_opengl.h>
/* screen width, height, and bit depth */
#define SCREEN_WIDTH  512
#define SCREEN_HEIGHT 512
#define SCREEN_BPP     16

/* This is our SDL surface */
SDL_Surface *surface;

GLfloat xrot=45;      /* X Rotation                           */
GLfloat yrot=70;      /* Y Rotation                           */
GLfloat z = -5.0f; /* Depth into the screen                */

/* Storage For Textures                 */
GLuint texture_image = 0;
GLuint texture_mask = 0;

/* Data we'll use to generate our cube */
GLfloat data[]= {
    /* Front Face */
    0.0f, 0.0f,             -1.0f, -1.0f, +1.0f,
    1.0f, 0.0f,             +1.0f, -1.0f, +1.0f,
    1.0f, 1.0f,             +1.0f, +1.0f, +1.0f,
    0.0f, 1.0f,             -1.0f, +1.0f, +1.0f,
    /* Back Face */
    1.0f, 0.0f,             -1.0f, -1.0f, -1.0f,
    1.0f, 1.0f,             -1.0f, +1.0f, -1.0f,
    0.0f, 1.0f,             +1.0f, +1.0f, -1.0f,
    0.0f, 0.0f,             +1.0f, -1.0f, -1.0f,
    /* Top Face */
    0.0f, 1.0f,             -1.0f, +1.0f, -1.0f,
    0.0f, 0.0f,             -1.0f, +1.0f, +1.0f,
    1.0f, 0.0f,             +1.0f, +1.0f, +1.0f,
    1.0f, 1.0f,             +1.0f, +1.0f, -1.0f,
    /* Bottom Face */
    1.0f, 1.0f,             -1.0f, -1.0f, -1.0f,
    0.0f, 1.0f,             +1.0f, -1.0f, -1.0f,
    0.0f, 0.0f,             +1.0f, -1.0f, +1.0f,
    1.0f, 0.0f,             -1.0f, -1.0f, +1.0f,
    /* Right Face */
    1.0f, 0.0f,             +1.0f, -1.0f, -1.0f,
    1.0f, 1.0f,             +1.0f, +1.0f, -1.0f,
    0.0f, 1.0f,             +1.0f, +1.0f, +1.0f,
    0.0f, 0.0f,             +1.0f, -1.0f, +1.0f,
    /* Left Face */
    0.0f, 0.0f,             -1.0f, -1.0f, -1.0f,
    1.0f, 0.0f,             -1.0f, -1.0f, +1.0f,
    1.0f, 1.0f,             -1.0f, +1.0f, +1.0f,
    0.0f, 1.0f,             -1.0f, +1.0f, -1.0f
};

// file reading stuff
const char *file2string(char *path)
{
    FILE *mf = fopen(path,"r");
    if(mf == NULL) { perror("opening file"); return NULL;}
    long len;
    fseek(mf,0,SEEK_END);
    len = ftell(mf);
    rewind(mf);
    char *str = malloc(len+1);
    if(str == NULL) { perror("mallocing"); return NULL;}
    fread(str,len,1,mf);
    str[len] = '\0';
    puts(str);
    return str;
}

// Shaders stuff
GLuint program = 0;
GLuint fragment_prog = 0;
GLuint vertex_prog = 0;
void setShaders()
{
    const char *vertex_shader_sources = NULL;
    const char *fragment_shader_sources = NULL;

    vertex_prog = glCreateShader(GL_VERTEX_SHADER);
    fragment_prog = glCreateShader(GL_FRAGMENT_SHADER);

    vertex_shader_sources = file2string("multitex.vert");
    fragment_shader_sources = file2string("multitex.frag");

    glShaderSource(vertex_prog, 1, &vertex_shader_sources,NULL);
    glShaderSource(fragment_prog, 1, &fragment_shader_sources,NULL);

    glCompileShader(vertex_prog);
    glCompileShader(fragment_prog);

    program = glCreateProgram();

    glAttachShader(program, vertex_prog);
    glAttachShader(program, fragment_prog);

    glLinkProgram(program);
    glUseProgram(program);

    free((char *)vertex_shader_sources);
    free((char *)fragment_shader_sources);
}
void unsetShaders()
{
    glDetachShader(program, fragment_prog);
    glDetachShader(program, vertex_prog);

    glDeleteShader(fragment_prog);
    glDeleteShader(vertex_prog);
    glDeleteProgram(program);
}

//Main stuff

/* function to load in bitmap as a GL texture */
int LoadGLTextures( )
{
    int status = 0;

    /* Create storage space for the texture */
    SDL_Surface *TextureImage;

    /* Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit */
    if ( ( TextureImage = SDL_LoadBMP( "base.bmp" ) ) )
    {
        glGenTextures( 1, &texture_image );
        glBindTexture( GL_TEXTURE_2D, texture_image );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

        /* Generate The Texture */
        glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, TextureImage->w,
                      TextureImage->h, 0, GL_BGR, GL_UNSIGNED_BYTE, TextureImage->pixels );

        /* Free up any memory we may have used */
        SDL_FreeSurface( TextureImage );
        status++;
    }

    if ( ( TextureImage = SDL_LoadBMP( "alpha.bmp" ) ) )
    {
        glGenTextures( 1, &texture_mask );
        glBindTexture( GL_TEXTURE_2D, texture_mask );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, TextureImage->w, TextureImage->h,
                             0, GL_ALPHA , GL_UNSIGNED_BYTE, TextureImage->pixels );
        SDL_FreeSurface( TextureImage );
        status++;
    }

    return status;
}

/* general OpenGL initialization function */
int initGL( GLvoid )
{
    /* Load in the texture */
    if ( !LoadGLTextures( ) )
        return 0;

    /* Enable Texture Mapping */
    glEnable( GL_TEXTURE_2D );

    glEnable(GL_BLEND);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    /* Enable smooth shading */
    glShadeModel( GL_SMOOTH );

    /* Set the background black */
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );

    /* Depth buffer setup */
    glClearDepth( 1.0f );

    /* Enables Depth Testing */
    glEnable( GL_DEPTH_TEST );

    /* The Type Of Depth Test To Do */
    glDepthFunc( GL_LEQUAL );

    /* Really Nice Perspective Calculations */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    glViewport( 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    gluPerspective( 45.0f, SCREEN_WIDTH / SCREEN_HEIGHT, 0.1f, 100.0f );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity( );

    return 1;
}

/* same as doMesh1TexelUnits except in 2 passes using 2 texel units */
void doMesh2TexelUnits( void )
{
    int i;

    /* TEXTURE-UNIT #0 */
    glActiveTexture( GL_TEXTURE0 );
    glBindTexture( GL_TEXTURE_2D, texture_image );

    /* TEXTURE-UNIT #1 */
    glActiveTexture( GL_TEXTURE1 );
    glBindTexture( GL_TEXTURE_2D, texture_mask );

    //Link textures to the corresponding uniform sampler2D in the fragment shader
    glUniform1i(glGetUniformLocation(program, "textureMap"), 0);
    glUniform1i(glGetUniformLocation(program, "alphaMap"), 1);

    glMatrixMode(GL_MODELVIEW); // go to our texture matrix
    glBegin( GL_QUADS );
    for ( i = 0; i < 24; i++ )
    {
        glMultiTexCoord2f( GL_TEXTURE0, data[5 * i], data[5 * i + 1] );
        glMultiTexCoord2f( GL_TEXTURE1, data[5 * i], data[5 * i + 1] );
        glVertex3f( data[5 * i + 2], data[5 * i + 3], data[5 * i + 4] );
    }
    glEnd();
}

/* Here goes our drawing code */
int drawGLScene( GLvoid )
{
    /* Clear The Screen And The Depth Buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glLoadIdentity( );
    glTranslatef( 0.0f, 0.0f, z );
    glRotatef( xrot, 1.0f, 0.0f, 0.0f );
    glRotatef( yrot, 0.0f, 1.0f, 0.0f );

    doMesh2TexelUnits( );

    /* Draw it to the screen */
    SDL_GL_SwapBuffers( );

    //~ xrot += 0.03f; /* X Axis Rotation */
    //~ yrot += 0.02f; /* Y Axis Rotation */
    //~ zrot += 0.04f; /* Z Axis Rotation */

    return 1;
}

int main( int argc, char **argv )
{
    /* Flags to pass to SDL_SetVideoMode */
    int videoFlags;
    /* main loop variable */
    int done = 0;
    /* used to collect events */
    SDL_Event event;
    /* this holds some info about our display */
    const SDL_VideoInfo *videoInfo;

    /* initialize SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
        {
            fprintf( stderr, "Video initialization failed: %s\n",
                     SDL_GetError( ) );
                exit(1);
        }

    /* Fetch the video info */
    videoInfo = SDL_GetVideoInfo( );

    if ( !videoInfo )
        {
            fprintf( stderr, "Video query failed: %s\n",
                     SDL_GetError( ) );
            exit(1);
        }

    /* the flags to pass to SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Enable OpenGL in SDL */
    videoFlags |= SDL_GL_DOUBLEBUFFER; /* Enable double buffering */
    videoFlags |= SDL_HWPALETTE;       /* Store the palette in hardware */

    /* This checks to see if surfaces can be stored in memory */
    if ( videoInfo->hw_available )
        videoFlags |= SDL_HWSURFACE;
    else
        videoFlags |= SDL_SWSURFACE;

    /* This checks if hardware blits can be done */
    if ( videoInfo->blit_hw )
        videoFlags |= SDL_HWACCEL;

    /* Sets up OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* get a SDL surface */
    surface = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
                                videoFlags );

    /* Verify there is a surface */
    if ( !surface )
    {
        fprintf( stderr,  "Video mode set failed: %s\n", SDL_GetError( ) );
        exit(1);
    }

    /* initialize OpenGL */
    glewInit();
    setShaders();
    initGL( );

    /* wait for events */
    while ( !done )
    {
        /* handle the events in the queue */
        while ( SDL_PollEvent( &event ) )
        {
            switch( event.type )
            {
                case SDL_KEYDOWN:
                    done = 1;
                    break;
                default:
                    break;
            }
        }
        drawGLScene( );
    }

    unsetShaders();

    /* clean up the window */
    SDL_Quit( );

    exit( EXIT_SUCCESS );
}
